from typing import Dict, Union, List

from aio_pika.abc import AbstractQueue

from .device_manager import RabbitMQDeviceManager


async def declare_queue(
        device_manager: RabbitMQDeviceManager,
        queue_name: str,
        passive: bool = False,
        arguments: Dict[str, Union[str, int, List[str]]] = None,
        durable: bool = True,
) -> AbstractQueue:
    return await (await device_manager.channel).declare_queue(
        name=queue_name,
        durable=durable,
        passive=passive,
        arguments=arguments,
    )



import random
from abc import ABC
from typing import Optional, List, Any, Dict

from aio_pika import connect_robust
from aio_pika.abc import AbstractRobustConnection, AbstractRobustChannel
from aio_pika.connection import make_url
from no_exception import NoException
from pamqp.constants import DEFAULT_PORT

from .base_device_manager import RabbitMQBaseDeviceManager


class RabbitMQDeviceManager(RabbitMQBaseDeviceManager, ABC):
    def __init__(
            self,
            hosts: List[str],
            user: str,
            password: str,
            vhost: str,
            publisher_confirms: bool,
            channel_qos_kwargs: Dict[str, Any] = None,
            use_transaction: bool = False,
            use_ssl: bool = False,
            port: int = DEFAULT_PORT,
    ):
        self._hosts = hosts
        self._user = user
        self._password = password
        self._vhost = vhost
        self._publisher_confirms = publisher_confirms
        self._channel_qos_kwargs = channel_qos_kwargs or {}
        self._use_transaction = use_transaction
        self._use_ssl = use_ssl
        self._port = port

        self._connection: Optional[AbstractRobustConnection] = None
        self._channel: Optional[AbstractRobustChannel] = None

    async def connect(self) -> None:
        if self._connection is None or self._connection.is_closed or \
                self._channel is None or self._channel.is_closed:
            await self._reconnect()

    async def close(self) -> None:
        await self._close_channel()
        await self._close_connection()

    @property
    async def connection(self) -> AbstractRobustConnection:
        if self._connection is None or self._connection.is_closed:
            await self._reconnect()
        return self._connection

    @property
    async def channel(self) -> AbstractRobustChannel:
        if self._channel is None or self._channel.is_closed:
            await self._reconnect()
        return self._channel

    async def _reconnect(self) -> None:
        if self._connection is None or self._connection.is_closed:
            await self._create_connection()
        if self._channel is None or self._channel.is_closed:
            await self._create_channel()

    async def _create_connection(self) -> None:
        self._connection = await connect_robust(
            make_url(
                host=random.choice(self._hosts),
                port=self._port,
                login=self._user,
                password=self._password,
                virtualhost=self._vhost,
                ssl=self._use_ssl,
            ),
        )

    async def _create_channel(self) -> None:
        self._channel = await self._connection.channel(
            self._publisher_confirms,
        )
        await self._channel.set_qos(
            **self._channel_qos_kwargs,
        )

    async def _close_connection(self) -> None:
        with NoException():
            await self._connection.close()

    async def _close_channel(self) -> None:
        with NoException():
            await self._channel.close()



from io import BytesIO
from typing import Optional, Tuple

from aio_pika.abc import HeadersType

from .base_device import RabbitMQBaseInputDevice
from .multi_connection_device_manager import RabbitMQMultiConnectionDeviceManager
from .transaction import BaseTransaction, RabbitMQIncomingMessageTransaction, EmptyTransaction


class RabbitMQInputBasicGetDevice(RabbitMQBaseInputDevice):
    def __init__(
            self,
            device_manager: RabbitMQMultiConnectionDeviceManager,
            device_name: str,
            use_transaction: bool,
    ):
        self._device_manager = device_manager
        self._device_name = device_name
        self._use_transaction = use_transaction

    async def commit_all_messages(self) -> None:
        raise NotImplemented("Not implemented for multi-connection")

    async def rollback_all_messages(self) -> None:
        raise NotImplemented("Not implemented for multi-connection")

    @property
    def has_messages_available(self) -> bool:
        return False

    async def read(
            self,
    ) -> Optional[Tuple[BytesIO, HeadersType, BaseTransaction]]:
        async with (await self._device_manager.channel).acquire() as channel:
            queue = await channel.get_queue(self._device_name)

            incoming_message = await queue.get(no_ack=not self._use_transaction, fail=False)
            if incoming_message is None:
                return None

            transaction = RabbitMQIncomingMessageTransaction(incoming_message) if self._use_transaction \
                else EmptyTransaction()

            return BytesIO(incoming_message.body), incoming_message.headers, transaction

    async def connect(self) -> None:
        pass

    async def close(self) -> None:
        pass



import asyncio
import threading
from asyncio import Future
from typing import List, TypeVar, Awaitable

from pamqp.common import Arguments
from pamqp.constants import DEFAULT_PORT

from .base_device_manager import RabbitMQBaseInputDeviceManager
from .device_manager import RabbitMQDeviceManager
from .input_consume_device import RabbitMQInputConsumeDevice

T = TypeVar('T')


class RabbitMQConsumeInputDeviceManager(
    RabbitMQDeviceManager,
    RabbitMQBaseInputDeviceManager,
):
    def __init__(
            self,
            hosts: List[str],
            user: str,
            password: str,
            vhost: str,
            prefetch_count: int,
            consumer_arguments: Arguments = None,
            use_transaction: bool = True,
            use_ssl: bool = False,
            port: int = DEFAULT_PORT,
    ):
        super().__init__(
            hosts=hosts,
            user=user,
            password=password,
            vhost=vhost,
            publisher_confirms=True,
            channel_qos_kwargs=dict(prefetch_count=prefetch_count),
            use_transaction=use_transaction,
            use_ssl=use_ssl,
            port=port,
        )
        self._consumer_arguments = consumer_arguments

        self._loop = asyncio.new_event_loop()
        self._thread_loop = threading.Thread(target=self._loop.run_forever)
        self._thread_loop.start()

    def run_in_executor(self, coro: Future[T] | Awaitable[T]) -> T:
        return asyncio.run_coroutine_threadsafe(coro, self._loop).result()

    async def get_device(
            self,
            device_name: str,
    ) -> RabbitMQInputConsumeDevice:
        return RabbitMQInputConsumeDevice(
            self,
            device_name,
            self._use_transaction,
            self._consumer_arguments,
        )
